# visualize.py

import tkinter as tk
import sys
from PIL import Image, ImageTk
import os
import warnings
from typing import (
    Iterable, Dict, Optional, Union, Callable, Any
)

import pandas as pd
import matplotlib.pyplot as plt

from market_break.process import up_movement, down_movement
from market_break.evaluation import results_report, EvaluationResults
from market_break.utils.base import assets

__all__ = [
    "EvaluationResultsWindow",
    "report_repr",
    "visualize_evaluation_results_plot",
    "visualize_evaluation_results_window"
]

if 'win' in sys.platform:
    from ctypes import windll

    app_id_stamp = (
        'ArbitrageScreener.application.Desktop-Application.1.0'
    )

    windll.user32.ShowWindow(
        windll.kernel32.GetConsoleWindow(), False
    )
    windll.shell32.SetCurrentProcessExplicitAppUserModelID(
        app_id_stamp
    )
# end if

plt.style.core.update_nested_dict(
    plt.style.library,
    plt.style.core.read_style_directory(
        f"{assets()}\\themes\\"
    )
)

def report_repr(
        data: Dict[str, Union[str, float]], title: Optional[str] = None
) -> str:
    """
    Returns a string representation of the results.

    :param data: The data to convert into a string.
    :param title: The title of the window.

    :return: The string representation of the data.
    """

    index = [0]

    for i, value in enumerate(data.values(), start=1):
        if value != "":
            index.append(index[-1] + 1)

        else:
            index.append(index[-1])
        # end if
    # end for

    return (
        "\n" + title + "\n\n" if isinstance(title, str) else ""
    ) + "\n".join(
        [
            (
                f"[{i}]{' ': <{3 - len(str(i))}}"
                f"{message: <35}" if value != '' else f'\n{message}'
            ) +
            f"{str(value if isinstance(value, str) else round(value, 3))}"
            for i, (message, value) in zip(index[1:], data.items())
        ]
    ) + "\n"
# end report_repr

class EvaluationResultsWindow(tk.Toplevel):
    """
    A class to represent a results window.

    The object to create and display a window to show the results
    of the simulation process.

    - results:
        The results object to display its data.

    - title:
        The title of the window to display.

    - theme:
        The theme of the window to display.

    - symbols:
        The symbols of the assets.

    - icon:
        The icon for the window.

    >>> from market_break.visualize import EvaluationResultsWindow
    >>>
    >>> EvaluationResultsWindow(...).mainloop()
    """

    parent = None

    def __init__(
            self,
            results: Optional[EvaluationResults] = None,
            title: Optional[str] = None,
            theme: Optional[str] = None,
            icon: Optional[str] = None,
            image: Optional[str] = None
    ) -> None:
        """
        Defines the class attributes.

        :param results: The results object.
        :param title: The title of the window.
        :param theme: The theme of the window.
        :param icon: The icon for the window.
        :param image: The path to the image file.
        """

        if EvaluationResultsWindow.parent is None:
            EvaluationResultsWindow.parent = tk.Tk()

            EvaluationResultsWindow.parent.withdraw()
        # end if

        super().__init__(EvaluationResultsWindow.parent)

        self.title_text: Optional[str] = title
        self.theme: Optional[str] = theme
        self.icon: Optional[str] = icon or f"{assets()}/icon/icon.ico"
        self.image: Optional[str] = image

        self.results: Optional[EvaluationResults] = results

        self.data: Dict[str, float] = {}

        self.build(
            results=results, title=title,
            theme=theme, icon=icon, image=image
        )
    # end __init__

    def __str__(self) -> str:
        """
        Returns a string representation of the object.

        :return: The string with the results.
        """

        return report_repr(
            data=self.data, title=self.title_text
        )
    # end __str__

    def build(
        self,
        results: EvaluationResults,
        title: Optional[str] = None,
        theme: Optional[str] = None,
        icon: Optional[str] = None,
        image: Optional[str] = None
    ) -> None:
        """
        Defines the class attributes.

        :param results: The results object.
        :param title: The title of the window.
        :param theme: The theme of the window.
        :param icon: The icon for the window.
        :param image: The path to the image file.
        """

        try:
            previous = self.winfo_children()

            if title:
                self.title_text = title
            # end if

            if theme:
                self.theme = theme
            # end if

            if icon:
                self.icon = icon
            # end if

            if image:
                self.image = image
            # end if

            title = title or self.title_text or f"Arbitrage Results"
            theme = theme or self.theme or "light"
            image = image or self.image

            self.title(title)
            self.iconbitmap(
                (
                    icon if (icon is not None and os.path.exists(icon)) else None
                ) or self.icon or f"{assets()}/icon/icon.ico"
            )

            self.results = results or self.results or EvaluationResults()

            if theme == "dark":
                bg = "#151515"
                fg = "#FFFFFF"

                self.config(bg=bg)

            elif theme == "light":
                fg = "#151515"
                bg = "#FFFFFF"

                self.config(bg=bg)

            else:
                bg = None
                fg = None
            # end if

            self.data = results_report(results=self.results)

            root = self

            frames = []

            for i, (message, value) in enumerate(self.data.items()):
                i: int

                if value == "":
                    root = tk.Frame(self, bg=bg)

                    root.grid_columnconfigure(0, weight=1)
                    root.grid_columnconfigure(1, weight=1)

                    frames.append(root)

                    tk.Label(
                        root, bg=bg, fg=fg, text=str(message),
                        font=("Ariel", 15, "bold")
                    ).grid(
                        row=i, column=0, sticky="nsw", padx=20, columnspan=2,
                        pady=(10 if i != 0 else (20, 10))
                    )

                else:
                    tk.Label(
                        root, bg=bg, fg=fg, text=str(message), font=("Ariel", 13)
                    ).grid(
                        row=i, column=0, sticky="nsw", padx=(20, 10),
                        pady=(5 if i != 0 else (10, 5))
                    )
                    tk.Label(
                        root, bg=bg, fg="#00b5fa", font=("Ariel", 13),
                        text=str(value if isinstance(value, (str, int)) else f"{value:.3f}")
                    ).grid(
                        row=i, column=1, sticky="nse", padx=(10, 20),
                        pady=(5 if i != len(self.data) - 1 else (5, 10))
                    )
                # end if
            # end for

            left_frames = frames[:len(frames) // 2 + 1 - len(frames) % 2]
            right_frames = frames[len(frames) // 2:]

            if len(left_frames) == 1:
                right_frames = left_frames + right_frames
                left_frames = []
            # end if

            if left_frames:
                self.grid_columnconfigure(0, weight=1)
            # end if

            if right_frames:
                self.grid_columnconfigure(1, weight=1)
            # end if

            for i, frame in enumerate(left_frames):
                frame.grid(
                    row=i, column=0, sticky="nsew",
                    padx=(15, 0), pady=(
                        (0, 15) if (i == len(left_frames) - 1) else 0
                    )
                )
            # end for

            for i in range(max(len(right_frames), len(left_frames))):
                self.grid_rowconfigure(i, weight=1)
            # end for

            for i, frame in enumerate(right_frames):
                frame.grid(
                    row=i, column=1, sticky="nsew",
                    padx=(15, 0), pady=(
                        (0, 15) if (i == len(right_frames) - 1) else 0
                    )
                )
            # end for

            if image and os.path.exists(image):
                position = 1

                if left_frames:
                    position += 1
                # end if

                if right_frames:
                    position += 1
                # end if

                image_object = ImageTk.PhotoImage(Image.open(image))

                image_label = tk.Label(self, image=image_object)
                image_label.image = image_object

                image_label.grid(
                    row=0, column=position, sticky="nsew", padx=20, pady=20,
                    rowspan=max(len(left_frames), len(right_frames))
                )

                self.columnconfigure(position, weight=1)
            # end if

            for child in previous:
                child.destroy()
            # end for

        except tk.TclError:
            pass
        # end try
    # end build
# end SimulationResultsWindow

def visualize_evaluation_results_window(
        results: Optional[EvaluationResults] = None,
        title: Optional[str] = None,
        theme: Optional[str] = None,
        icon: Optional[str] = None
) -> EvaluationResultsWindow:
    """
    Defines the class attributes.

    :param results: The results object.
    :param title: The title of the window.
    :param theme: The theme of the window.
    :param icon: The icon for the window.
    """

    return EvaluationResultsWindow(
        results=results, title=title,
        theme=theme, icon=icon
    )
# end visualize_evaluation_results_window

def visualize_evaluation_results_plot(
        results: EvaluationResults,
        save: Optional[str] = None,
        additions: Optional[
            Union[
                Iterable[Callable[[], Any]],
                Dict[str, Iterable[float]],
                Iterable[Iterable[float]],
                bool
            ]
        ] = None,
        show: Optional[bool] = True,
        title: Optional[str] = None,
        trade_index: Optional[Iterable] = None,
        icon: Optional[str] = None,
        theme: Optional[str] = None,
) -> plt.Figure:
    """
    Visualizes the process results.

    :param results: The results object.
    :param save: The value to save the image.
    :param show: The value to show the image.
    :param theme: The theme of the window.
    :param additions: The additions to run to add to the plot.
    :param title: The title of the window.
    :param trade_index: The index for the plot.
    :param icon: The path to an icon.
    """

    icon = icon or f"{assets()}/icons/icon.ico"
    theme = theme or "light"

    additions = additions or {}
    commands = {}

    if theme:
        plt.style.use(theme)
    # end if

    warnings.resetwarnings()

    if isinstance(additions, dict):
        additions.update(commands)

    else:
        additions = [*additions, *commands.values()]
    # end if

    with warnings.catch_warnings():
        warnings.filterwarnings("ignore")
        plot = plt.figure(
            "Arbitrage Trading",
            figsize=(13 if additions else 10, 7.5 if additions else 6)
        )

    plt.subplots_adjust(
        top=0.938, bottom=0.067, right=0.974,
        left=0.06, hspace=0.283, wspace=0
    )

    labeled = False

    if additions:
        plt.subplot(2, 1, 1)
        plt.title(title or f"Arbitrage Trading Performance")

        for addition in additions:
            if (addition is not None) and callable(addition):
                addition()

            elif addition is not None:
                if isinstance(additions, dict):
                    if not callable(additions[addition]):
                        labeled = True

                        plt.plot(
                            (
                                list(range(len(additions[addition])))
                                if not isinstance(additions[addition], pd.Series)
                                else additions[addition].index
                            ),
                            additions[addition],
                            lw=1.25, label=addition
                        )

                    else:
                        additions[addition]()
                    # end if

                elif isinstance(addition, pd.Series):
                    labeled = True

                    plt.plot(
                        (
                            list(range(len(additions[addition])))
                            if not isinstance(addition, pd.Series)
                            else addition.index
                        ),
                        addition, lw=1.25, label=addition.name
                    )

                else:
                    plt.plot(list(range(len(addition))), addition, lw=1.25)
                # end if
            # end if
        # end for

        plt.xlabel("Time")
        plt.ylabel("Normalized Price Spread")

        if labeled:
            plt.legend()
        # end if
    # end if

    trade_index = trade_index or range(len(results.balances))

    plt.subplot(2 if additions else 1, 1, 2 if additions else 1)
    plt.title(
        title or (
            "Arbitrage Trading Results "
            f"({(results.balances[-1] / results.balances[0] - 1) * 100:.3f}%)"
        )
    )

    predictions = None

    if (results.predictions is not None) and (len(results.predictions) > 0):
        predictions = ((results.predictions / results.predictions[0]) - 1) * 100
    # end if

    balances = (results.balances / results.balances[0] - 1) * 100

    profits = up_movement(balances)
    losses = -down_movement(balances)

    if predictions is not None:
        plt.plot(trade_index, predictions, lw=1.25, label=f'Predicted Balance', color='blue')
        plt.fill_between(trade_index, predictions, balances, color="blue", alpha=0.1)
        plt.fill_between(trade_index, balances, predictions, color="blue", alpha=0.1)
    # end if

    plt.plot(trade_index, balances, lw=1.25, label=f'Balance', color='orange')
    plt.plot(trade_index, profits, lw=1.25, label=f'Profits', color='green')
    plt.plot(trade_index, losses, lw=1.25, label=f'Losses', color='red')

    plt.fill_between(trade_index, profits, balances, color="green", alpha=0.1)
    plt.fill_between(trade_index, balances, losses, color="yellow", alpha=0.1)
    plt.fill_between(trade_index, losses, color="red", alpha=0.1)
    plt.xlabel("Time")
    plt.ylabel("Produced Balance (%)")
    plt.legend()

    if save:
        if os.path.split(save)[0]:
            os.makedirs(os.path.split(save)[0], exist_ok=True)
        # end if

        plt.savefig(save)
    # end if

    if isinstance(icon, str) and os.path.exists(icon):
        manager = plt.get_current_fig_manager()

        # noinspection PyProtectedMember
        manager.window.tk.call('wm', 'iconbitmap', manager.window._w, icon)
    # end if

    warnings.resetwarnings()

    if show:
        plt.show()
    # end if

    return plot
# end visualize_evaluation_results_plot


# dataset.py

import os
import datetime as dt
import json
from pathlib import Path
from typing import (
    Union, Optional, Tuple, Iterable,
    Any, Callable, Dict, List
)

import numpy as np
import pandas as pd
from pandas.core.resample import Resampler

from market_break.interval import interval_total_time

__all__ = [
    "row_to_dataset",
    "save_dataset",
    "load_dataset",
    "update_dataset",
    "split_dataset",
    "strip_dataset",
    "validate_dataset",
    "OHLCV_COLUMNS",
    "OHLC_COLUMNS",
    "OPEN",
    "CLOSE",
    "HIGH",
    "LOW",
    "VOLUME",
    "DATE_TIME",
    "BIDS",
    "ASKS",
    "find_column",
    "validate_file_extension",
    "dataset_to_json",
    "dataset_from_json",
    "EXTENSIONS",
    "validate_file_path",
    "is_valid_location",
    "prepare_saving_location",
    "BIDS_VOLUME",
    "ASKS_VOLUME",
    "CSV_EXTENSION",
    "JSON_EXTENSION",
    "DEFAULT_EXTENSION",
    "bid_ask_to_ohlcv",
    "create_dataset",
    "ORDERBOOK_COLUMNS",
    "adjust_series",
    "interval_adjuster",
    "index_to_datetime",
    "new_dataset_index"
]

DATE_TIME = 'DateTime'

OPEN = "Open"
CLOSE = "Close"
HIGH = "High"
LOW = "Low"
VOLUME = "Volume"

BIDS = "Bids"
ASKS = "Asks"
BIDS_VOLUME = "BidsVolume"
ASKS_VOLUME = "AsksVolume"

OHLC_COLUMNS = (OPEN, HIGH, LOW, CLOSE)
OHLCV_COLUMNS = (*OHLC_COLUMNS, VOLUME)
ORDERBOOK_COLUMNS = (BIDS, ASKS, BIDS_VOLUME, ASKS_VOLUME)

def index_to_datetime(index: Any, adjust: Optional[bool] = True) -> dt.datetime:
    """
    Converts the index into a datetime object.

    :param index: The value to convert.
    :param adjust: The value to adjust the process for errors.

    :return: The datetime object.
    """

    try:
        if isinstance(index, str):
            index = dt.datetime.fromisoformat(index)

        elif isinstance(index, (int, float)):
            index = dt.datetime.fromtimestamp(index)

        elif isinstance(index, pd.Timestamp):
            index = index.to_pydatetime()

        elif isinstance(index, np.datetime64):
            index = index.astype(dt.datetime)
        # end if

    except (TypeError, ValueError) as e:
        if adjust:
            pass

        else:
            raise e
        # end if
    # end try

    return index
# end index_to_datetime

def row_to_dataset(
        dataset: Union[pd.DataFrame, pd.Series],
        index: Optional[int] = None
) -> pd.DataFrame:
    """
    Creates a dataframe from the row.

    :param dataset: The base dataset from witch the row came.
    :param index: The index of the row to create a dataset for.

    :return: The dataset from the row.
    """

    if isinstance(dataset, pd.DataFrame):
        if index is None:
            raise ValueError(
                f"Index must an int when dataset "
                f"is of type {pd.DataFrame}."
            )
        # end if

        return pd.DataFrame(
            {
                column: [value] for column, value in
                dict(dataset.iloc[index]).items()
            },
            index=[dataset.index[index]]
        )

    elif isinstance(dataset, pd.Series):
        return pd.DataFrame(
            {
                column: [value] for column, value in
                dict(dataset).items()
            },
            index=[index or 0]
        )

    else:
        raise TypeError(
            f"Dataset must be either of type {pd.DataFrame}, "
            f"or {pd.Series}, not {type(dataset)}."
        )
    # end if
# end row_to_dataset

def update_dataset(base: pd.DataFrame, new: pd.DataFrame) -> None:
    """
    Updates the ba se dataframe with new columns from the new dataframe.

    :param base: The base dataframe to update.
    :param new: The new dataframe with the new columns.
    """

    if not len(base) == len(new):
        raise ValueError(
            f"DataFrames lengths must match "
            f"(got {len(base)} and {len(new)} instead)."
        )
    # end if

    for column in new.columns:
        if column not in base.columns:
            base[column] = new[column]
        # end if
    # end for
# end update_dataset

def split_dataset(
        dataset: Union[pd.DataFrame, pd.Series],
        size: Optional[Union[int, float]] = None,
        length: Optional[int] = None
) -> Tuple[pd.DataFrame, pd.DataFrame]:
    """
    Splits the new_dataset into to parts at the point of the given size.

    :param dataset: The new_dataset to split.
    :param size: The size of the first part.
    :param length: The length of the split.

    :return: The two datasets.
    """

    if (size is None) and (length is None):
        raise ValueError(
            "Cannot split the dataset when neither "
            "size nor length parameters are defined."
        )
    # end if

    length = length or int(len(dataset) * size)

    return dataset[:length], dataset[length:]
# end split_dataset

def strip_dataset(dataset: pd.DataFrame, columns: Iterable[str]) -> pd.DataFrame:
    """
    Strips the columns from the new_dataset.

    :param dataset: The new_dataset to remove features from.
    :param columns: The columns to validate.

    :return: The new new_dataset.
    """

    return dataset.drop(
        [
            column for column in columns
            if column in dataset.columns
        ], axis=1
    )
# end strip_dataset

def validate_dataset(
        dataset: pd.DataFrame,
        columns: Optional[Iterable[str]] = None,
        length: Optional[int] = None
) -> None:
    """
    Validates the new_dataset to have the columns.

    :param dataset: The new_dataset to validate.
    :param columns: The columns to validate.
    :param length: The length of the valid dataset.
    """

    if (
        (columns is not None) and
        not all(column in dataset.columns for column in columns)
    ):
        missing = [
            column for column in columns
            if column not in list(dataset.columns)
        ]

        redundant = [
            column for column in list(dataset.columns)
            if column not in columns
        ]

        raise ValueError(
            f"DataFrame must include the "
            f"columns by the names: {columns}.\n"
            f"Given columns: {', '.join(dataset.columns)}.\n"
            f"Missing columns: {missing}.\n"
            f"Redundant columns: {redundant}."
        )
    # end if

    if (length is not None) and len(dataset) != length:
        raise ValueError(
            f"Dataset must have length of {length}, "
            f"not: {len(dataset)}."
        )
    # end if
# end validate_dataset

def dataset_to_json(dataset: pd.DataFrame) -> List[Union[str, Dict[str, Any]]]:
    """
    Converts the data of the dataset to json.

    :param dataset: The dataset to process.

    :return: The json representation of the data.
    """

    return list(json.loads(dataset.to_json(orient='index')).items())
# end dataset_to_json

def dataset_from_json(
        data: Union[Dict[str, Dict[str, Any]], List[Union[str, Dict[str, Any]]]]
) -> pd.DataFrame:
    """
    Converts the data from json format into a dataframe object.

    :param data: The json data to process.

    :return: The data frame object.
    """

    if isinstance(data, list):
        data = dict(data)
    # end if

    return pd.read_json(json.dumps(data), orient="index")
# end dataset_from_json

CSV_EXTENSION = "csv"
JSON_EXTENSION = "json"
DEFAULT_EXTENSION = CSV_EXTENSION

EXTENSIONS = (CSV_EXTENSION, JSON_EXTENSION)

def validate_file_extension(
        path: Union[str, Path],
        extension: Optional[str] = None
) -> str:
    """
    Validates the file formatting.

    :param path: The path to the file.
    :param extension: The data formatting.

    :return: The valid formatting.
    """

    path = str(path)

    if extension is None:
        if "." not in path:
            raise ValueError(
                f"Cannot infer file type and data "
                f"format from path: {path} and undefined formatting. "
                f"You may need to specify file extension in the path "
                f"or pass the 'formatting' parameter ({', '.join(EXTENSIONS)})."
            )
        # end if

        extension = path[path.rfind(".") + 1:]
    # end if

    if extension not in EXTENSIONS:
        raise ValueError(
            f"Invalid formatting value: {extension}. "
            f"value formatting options are: {', '.join(EXTENSIONS)}."
        )
    # end if

    return extension
# end validate_file_extension

def is_valid_location(path: Union[str, Path]) -> bool:
    """
    Prepares the saving location.

    :param path: The path for the file to save.

    :return: The value of creating the location directory.
    """

    location = os.path.split(path)[0]

    return (
        ((not location) and path) or
        (location and os.path.exists(location))
    )
# end is_valid_location

def validate_file_path(
        path: Union[str, Path],
        create: Optional[bool] = True,
        override: Optional[bool] = True
) -> None:
    """
    Validates the file formatting.

    :param path: The path to the file.
    :param create: The value to create the path location.
    :param override: The value to override an existing file.

    :return: The valid formatting.
    """

    if create:
        prepare_saving_location(path=path)

    elif not is_valid_location(path=path):
        raise ValueError(
            f"Invalid file saving "
            f"location: {os.path.split(path)[0]} of {path}."
        )
    # end if

    if os.path.exists(path) and not override:
        raise FileExistsError(
            f"Attempting to override an existing file: "
            f"{path} while 'override' is set to {override}."
        )
    # end if
# end validate_file_path

def prepare_saving_location(path: Union[str, Path]) -> bool:
    """
    Prepares the saving location.

    :param path: The path for the file to save.

    :return: The value of creating the location directory.
    """

    location = os.path.split(path)[0]

    if location:
        value = os.path.exists(location)

        os.makedirs(location, exist_ok=True)

        return value

    else:
        return False
    # end if
# end prepare_saving_location

def save_dataset(
        dataset: pd.DataFrame,
        path: Union[str, Path],
        headers: Optional[bool] = None,
        create: Optional[bool] = True,
        override: Optional[bool] = True,
        append: Optional[bool] = False,
        extension: Optional[str] = None
) -> None:
    """
    Saves the data.

    :param dataset: The dataset to save.
    :param create: The value to create the path location.
    :param headers: The value to include headers.
    :param override: The value to override an existing file.
    :param path: The saving path.
    :param append: The value to append data to the file.
    :param extension: The formatting of the data.
    """

    if extension is None:
        extension = DEFAULT_EXTENSION
    # end if

    if headers is None:
        headers = not append
    # end if

    path = str(path)

    extension = validate_file_extension(path=path, extension=extension)

    validate_file_path(path=path, create=create, override=override)

    if extension == CSV_EXTENSION:
        dataset.to_csv(path, mode='a' if append else "w", header=headers)

    elif extension == JSON_EXTENSION:
        with open(path, "w") as file:
            json.dump(dataset_to_json(dataset), file)
        # end open
    # end if
# end save_dataset

def load_dataset(
        path: Union[str, Path],
        extension: Optional[str] = None,
        index_column: Optional[Union[int, bool]] = 0,
        time_index: Optional[bool] = True
) -> pd.DataFrame:
    """
    Loads the dataset from the path.

    :param path: The saving path.
    :param extension: The formatting of the data.
    :param index_column: The value to set the index for the column.
    :param time_index: The value to se the index as datetime.

    :return: The loaded dataset.
    """

    if extension is None:
        extension = DEFAULT_EXTENSION
    # end if

    path = str(path)

    extension = validate_file_extension(path=path, extension=extension)

    if extension == CSV_EXTENSION:
        dataset = pd.read_csv(path)

    elif extension == JSON_EXTENSION:
        with open(path, "r") as file:
            dataset = dataset_to_json(json.load(file))
        # end open
    # end if

    if index_column is True:
        index_column = 0
    # end if

    if index_column is not None or index_column is False:
        index_column_name = list(dataset.columns)[index_column]
        dataset.index = (
            pd.DatetimeIndex(dataset[index_column_name])
            if time_index else dataset[index_column_name]
        )
        del dataset[index_column_name]
        dataset.index.name = DATE_TIME
    # end if

    return dataset
# end load_dataset

def find_column(
        dataset: pd.DataFrame,
        columns: Iterable[Any],
        validation: Optional[Callable[[pd.Series], bool]] = None
) -> Optional[pd.Series]:
    """
    Finds the first valid column and returns it.

    :param dataset: The dataset to search.
    :param columns: The column names to search from, by order.
    :param validation: The validation function.

    :return: The valid column.
    """

    for column in columns:
        if column not in dataset:
            continue
        # end if

        if (
            (validation is None) or
            (callable(validation) and validation(dataset[column]))
        ):
            return dataset[column]
        # end if
    # end for
# end find_column

def interval_adjuster(interval: str) -> str:
    """
    Creates the adjuster for the interval.

    :param interval: The interval to adjust data with.

    :return: The adjusted interval.
    """

    return f'{interval_total_time(interval).total_seconds()}S'
# end interval_adjuster

def adjust_series(data: pd.Series, interval: str) -> Resampler:
    """
    Converts the dataset into a dataset with an interval.

    :param data: The source data.
    :param interval: The interval for the new dataset.

    :return: The returned dataset.
    """

    return data.resample(interval_adjuster(interval))
# end adjust_dataset

def bid_ask_to_ohlcv(dataset: pd.DataFrame, interval: str) -> pd.DataFrame:
    """
    Converts the BID/ASK spread dataset into a OHLCV dataset.

    :param dataset: The source data.
    :param interval: The interval for the new dataset.

    :return: The returned dataset.
    """

    if not all(column in ORDERBOOK_COLUMNS for column in dataset.columns):
        raise ValueError(
            f"Dataset has to contain all columns: "
            f"{', '.join(ORDERBOOK_COLUMNS)}, "
            f"but found only: {', '.join(dataset.columns)}"
        )
    # end if

    adjuster = interval_adjuster(interval)

    ohlcv_dataset = dataset[BIDS].resample(adjuster).ohlc()
    ohlcv_dataset.columns = list(OHLC_COLUMNS)
    ohlcv_dataset[VOLUME] = dataset[BIDS_VOLUME].resample(adjuster).mean()

    return ohlcv_dataset
# end bid_ask_to_ohlcv

def create_dataset(columns: Optional[Iterable[str]] = None) -> pd.DataFrame:
    """
    Creates a dataframe for the order book data.

    :param columns: The dataset columns.

    :return: The dataframe.
    """

    market = pd.DataFrame(
        {column: [] for column in columns or []}, index=[]
    )
    market.index.name = DATE_TIME

    return market
# end create_dataset

def new_dataset_index(index: dt.datetime, dataset: pd.DataFrame) -> bool:
    """
    Checks if the index is new and valid to the dataset.

    :param index: The index to check.
    :param dataset: The dataset to validate the index for.

    :return: The validation of the index as new.
    """

    return (
        (len(dataset.index) == 0) or
        (
            isinstance(dataset.index[-1], dt.datetime) and
            (dataset.index[-1] < index)
        ) or
        (index not in dataset.index)
    )
# end new_dataset_index


# base.py

import os
from pathlib import Path

__all__ = [
    "root",
    "source",
    "assets"
]

def root() -> str:
    """
    Returns the root of the source program.

    :return: The path to the source.
    """

    try:
        if os.getcwd() in os.environ['VIRTUAL_ENV']:
            path = Path(__file__).parent.parent

        else:
            raise KeyError
        # end if

    except KeyError:
        if os.getcwd() not in (
            path := str(Path(__file__).parent.parent)
        ):
            path = os.getcwd()
        # end if
    # end try

    return str(path)
# end root

def source() -> str:
    """
    Returns the root of the source program.

    :return: The path to the source.
    """

    return str(Path(root()) / Path("source"))
# end source

def assets() -> str:
    """
    Returns the root of the source program.

    :return: The path to the source.
    """

    return str(Path(source()) / Path("assets"))
# end assets


# experiment.py

from typing import Iterable, List, Tuple, Optional

from attrs import define

import pandas as pd
import numpy as np

from market_break.process import array

__all__ = [
    "buy_sell_signals",
    "long_short_results",
    "TradingRecord",
    "ShortTradingRecord",
    "LongTradingRecord"
]

def buy_sell_signals(
        data: Iterable[float], up: Iterable[int], down: Iterable[int]
) -> Tuple[List[int], List[int]]:
    """
    Calculates the buy and sell signals from the trends.

    :param data: The data for the indicator.
    :param up: The up-trends indexes.
    :param down: The down-trends indexes.

    :return: The buy and sell indexes.
    """

    buy = []
    sell = []

    for i in range(len(pd.Series(data))):
        if (i in up) and (len(buy) == len(sell)):
            buy.append(i)

        elif (i in down) and (len(buy) > len(sell)):
            sell.append(i)
        # end if
    # end for

    return buy, sell
# end buy_sell_signals

@define
class TradingRecord:
    """A class to represent the trading record."""

    returns: np.ndarray
    entries: np.ndarray
    exits: np.ndarray
    data: Iterable[float]
    buy: Iterable[int]
    sell: Iterable[int]
    long: Optional[bool] = True
    short: Optional[bool] = True
    volume: Optional[float] = 1.0
    fee: Optional[float] = 0.0
    take: Optional[float] = None
    stop: Optional[float] = None
# end TradingRecord

class LongTradingRecord(TradingRecord):
    """A class to represent the trading record."""
# end LongTradingRecord

class ShortTradingRecord(TradingRecord):
    """A class to represent the trading record."""
# end ShortTradingRecord

def difference(
        current: float, previous: float, fee: Optional[float] = 0.0
) -> float:
    """
    Calculates the relative difference between the values, with fee.

    :param current: The curren value.
    :param previous: The last value.
    :param fee: The fee to take off.

    :return: The difference.
    """

    return ((current - previous) / previous) * (1 - fee)
# end difference

def long_short_results(
        data: Iterable[float],
        buy: Iterable[int],
        sell: Iterable[int],
        volume: Optional[float] = 1.0,
        fee: Optional[float] = 0.0,
        take: Optional[float] = None,
        stop: Optional[float] = None
) -> Tuple[LongTradingRecord, ShortTradingRecord]:
    """
    Calculates the long and short signals from the indexes.

    :param data: The data for the indicator.
    :param buy: The buy indexes.
    :param sell: The sell indexes.
    :param volume: The initial investment volume.
    :param fee: The fee for an action.
    :param take: The take profit threshold.
    :param stop: The stop loss threshold.

    :return: The long and short indexes.
    """

    long_returns = [volume]
    short_returns = [volume]

    is_long = False
    is_short = False

    last_long = 0
    last_short = 0

    long_entries = []
    short_entries = []
    long_exits = []
    short_exits = []

    data = array(data)

    for i in range(1, len(data)):
        diff = difference(current=data[i], previous=data[i - 1], fee=fee)

        if is_long and (
            ((take is not None) and (last_long * take <= data[i])) or
            ((stop is not None) and (last_long * (1 - stop) >= data[i]))
        ):
            is_long = False

            long_exits.append(i)
            long_returns.append(long_returns[-1] + diff)
            short_returns.append(short_returns[-1])

            continue
        # end if

        if is_short and (
            ((take is not None) and (last_short * take >= data[i])) or
            ((stop is not None) and (last_short * (1 - stop) <= data[i]))
        ):
            is_short = False

            short_exits.append(i)
            short_returns.append(short_returns[-1] - diff)
            long_returns.append(long_returns[-1])

            continue
        # end if

        if is_long or (i in buy):
            is_short = False

            short_exits.append(i)
            long_returns.append(long_returns[-1] + diff)

            if not is_long:
                long_entries.append(i)

                is_long = True
                last_long = data[i]
            # end if

        else:
            long_returns.append(long_returns[-1])
        # end if

        if is_short or (i in sell):
            is_long = False

            long_exits.append(i)
            short_returns.append(short_returns[-1] - diff)

            if not is_short:
                short_entries.append(i)

                is_short = True
                last_short = data[i]
            # end if

        else:
            short_returns.append(short_returns[-1])
        # end if
    # end for

    return (
        LongTradingRecord(
            returns=np.array(long_returns).astype(float),
            entries=np.array(long_entries).astype(int),
            exits=np.array(long_exits).astype(int),
            data=data, buy=buy, sell=sell, volume=volume,
            fee=fee, take=take, stop=stop
        ),
        ShortTradingRecord(
            returns=np.array(short_returns).astype(float),
            entries=np.array(short_entries).astype(int),
            exits=np.array(short_exits).astype(int),
            data=data, buy=buy, sell=sell, volume=volume,
            fee=fee, take=take, stop=stop
        )
    )
# end long_short_results


# process.py

import math
from typing import (
    Iterable, Optional, Callable,
    Any, Sequence, Tuple, List, Union
)

import numpy as np
import pandas as pd

__all__ = [
    "average_deviation_rate",
    "average_growth_rate",
    "average_win_rate",
    "average_shrink_rate",
    "average_gain_rate",
    "round_decimals_down",
    "weighted_average_win_rate",
    "average_change_rate",
    "cumulative_shrink",
    "cumulative_gain",
    "normalize_calculation",
    "validate_data_length",
    "minimums",
    "maximums",
    "adjust_peaks",
    "peaks",
    "normalize",
    "array",
    "linear_distance_clusters",
    "drop_nan",
    "up_movement",
    "down_movement",
    "spread_deviation_risk_ratio"
]

np.seterr(divide='ignore', invalid='ignore')

def validate_data_length(
        data: Sequence,
        minimum: Optional[int] = None,
        maximum: Optional[int] = None
) -> None:
    """
    Validates the length of the data.

    :param data: The data to validate.
    :param minimum: The minimum length.
    :param maximum: The maximum length.
    """

    length = len(data)

    if minimum is None:
        minimum = 0
    # end if

    if (minimum is not None) and (length < minimum):
        raise ValueError(
            f"Minimum length of {data} is {minimum}, "
            f"but given length is {length}."
        )
    # end if

    if (maximum is not None) and (length > maximum):
        raise ValueError(
            f"Maximum length of {data} is {maximum}, "
            f"but given length is {length}."
        )
    # end if
# end validate_data_length

def minimums(data: Iterable[float]) -> np.array:
    """
    Returns an array containing the indexes of the local minimum values.

    :param data: The data to process.

    :return: An array of indexes.
    """

    return (np.diff(np.sign(np.diff(array(data)))) > 0).nonzero()[0] + 1
# end minimums

def maximums(data: Iterable[float]) -> np.array:
    """
    Returns an array containing the indexes of the local maximum values.

    :param data: The data to process.

    :return: An array of indexes.
    """

    return (np.diff(np.sign(np.diff(array(data)))) < 0).nonzero()[0] + 1
# end maximums

def adjust_peaks(
        buy: Iterable[float], sell: Iterable[float]
) -> Tuple[np.array, np.array]:
    """
    Processes the minimum and maximum peaks to match the trading process.

    :param buy: The indexes of buying-actions.
    :param sell: The indexes of selling-actions.

    :return: An array of the minimum values indexes and an array of the maximum values indexes.
    """

    buy = array(buy)
    sell = array(sell)

    if (sell.size > 0) and (buy.size > 0) and (sell[0] < buy[0]):
        sell = sell[1:]
    # end if

    return buy, sell
# end peaks

def peaks(
        data: Iterable[float], adjust: Optional[bool] = False
) -> Tuple[np.array, np.array]:
    """
    Processes the minimum and maximum peaks to match the trading process.

    :param data: The data to process.
    :param adjust: Thw value to adjust the peaks.

    :return: An array of the minimum values indexes and an array of the maximum values indexes.
    """

    data = array(data)

    maxima = maximums(data)
    minima = minimums(data)

    if adjust:
        return adjust_peaks(minima, maxima)
    # end if

    return minima, maxima
# end peaks

def normalize(data: Iterable[float]) -> np.array:
    """
    Normalizes the values between 0 and 1.

    :param data: The data to normalize.

    :return: The normalized array of values.
    """

    if not isinstance(data, np.ndarray):
        data = np.array(data)
    # end if

    return (data - np.min(data)) / np.ptp(data)
# end normalize

def array(data: Iterable[float], scale: Optional[bool] = False) -> np.array:
    """
    Normalizes the values between 0 and 1.

    :param data: The data to normalize.
    :param scale: The valur to normalize the data.

    :return: The normalized array of values.
    """

    if not isinstance(data, np.ndarray):
        data = np.array(data)
    # end if

    if scale and not ((np.max(data) == 1) and (np.min(data) == 0)):
        data = normalize(data)
    # end if

    return data
# end normalize

def drop_nan(data: Iterable[float], value: Optional = 0.0) -> np.array:
    """
    Replaces the nan values in the data with a given value.

    :param data: The data to refactor.
    :param value: The value to replace nan values.

    :return: The refactored data.
    """

    return np.nan_to_num(array(data), nan=value)
# end drop_nan

def linear_distance_clusters(
        data: Iterable[Union[float, np.ndarray]],
        index: Optional[Iterable] = None,
        source: Optional[Union[float, np.array]] = None
) -> Tuple[List[List], List[List]]:
    """
    Clusters the data into clusters, with small distance between each element in the cluster.

    :param data: The data to cluster by linear distance.
    :param index: The index of the data.
    :param source: The relative source for the data.

    :return: The clusters of the data index.
    """

    data = array(data)

    if data.size == 0:
        raise ValueError(f"Empty data sequence given.")
    # end if

    if index is None:
        index = range(data.size)
    # end if

    dataset = list(zip(index, data))

    dataset.sort(
        key=lambda pair: (
            (np.linalg.norm(pair[1] - source)
             if (source is not None) else
             np.linalg.norm(pair[1]))
        )
    )

    indexes = [pair[0] for pair in dataset]
    data = [pair[1] for pair in dataset]

    data_clusters = [[data[0]]]
    index_clusters = [[indexes[0]]]

    if len(data) == 1:
        return index_clusters, data_clusters
    # end if

    for i, value in enumerate(data[1:], start=1):
        if i in index_clusters[-1]:
            continue
        # end if

        if (
            (len(data_clusters[-1]) >= 2) and
            (
                np.linalg.norm(value - data_clusters[-1][-1]) >
                np.linalg.norm(data_clusters[-1][-1] - data_clusters[-1][-2])
            )
        ):
            if (i + 1 < len(data)) and (
                np.linalg.norm(data[i + 1] - value) <
                np.linalg.norm(value - data_clusters[-1][-1])
            ):
                data_clusters.append([value, data[i + 1]])
                index_clusters.append([indexes[i], indexes[i + 1]])

            else:
                data_clusters.append([value])
                index_clusters.append([indexes[i]])
            # end if
        # end if

        else:
            data_clusters[-1].append(value)
            index_clusters[-1].append(indexes[i])
        # end if
    # end for

    return index_clusters, data_clusters
# end linear_distance_clusters

def normalize_calculation(calculation: Callable) -> Callable:
    """
    Creates a decorator to normalize the mathematical calculation.

    :param calculation: The calculation function.

    :return: The wrapper function to wrap the calculation.
    """

    def wrapper(*args: Any, **kwargs: Any) -> Any:
        """
        Runs the given calculation with its parameters.

        :param args: The arguments for the calculation.
        :param kwargs: The keyword arguments for the calculation.

        :return: The returned results from the calculation.
        """

        try:
            return calculation(*args, **kwargs)

        except (ValueError, ZeroDivisionError):
            return 0
        # end try
    # end wrapper

    return wrapper
# end normalize_calculation

def round_decimals_down(number: float, decimals: Optional[int] = 0) -> float:
    """
    Returns a value rounded down to a specific number of decimal places.

    :param number: The number to round
    :param decimals: The amount of decimal places to leave in the number.

    :return: The rounded value.

    """

    if not isinstance(decimals, int):
        raise TypeError("Decimal places must be an integer.")

    elif decimals < 0:
        raise ValueError("Decimal places has to be 0 or more.")

    elif decimals == 0:
        return float(math.floor(number))
    # end if

    string_number = str(float(number))

    return float(string_number[:string_number.find(".") + decimals + 1])
# end round_decimals_down

def average_growth_rate(data: Iterable[float]) -> float:
    """
    Calculates the average change rate.

    :param data: The simulation data.

    :return: The calculated number.
    """

    data = array(data)

    validate_data_length(data)

    try:
        change = data[1:] - pd.Series(data).shift(1)[1:]

        if change.size == 0:
            return 0
        # end if

        s = np.sum(change)

        if np.isnan(s):
            return 0
        # end if

        return np.sum(s) / change.size

    except (ValueError, ZeroDivisionError, IndexError):
        return 0
    # end try
# end average_growth_rate

def average_shrink_rate(data: Iterable[float]) -> float:
    """
    Calculates the average change rate.

    :param data: The simulation data.

    :return: The calculated number.
    """

    data = array(data)

    validate_data_length(data)

    try:
        change = data[1:] - pd.Series(data).shift(1)[1:]

        change = np.where(change < 0, change, 0)
        change = np.array(
            [np.sum(change[:i + 1]) for i in range(change.size)]
        )
        change = change[1:] - pd.Series(change).shift(1)[1:]

        if change.size == 0:
            return 0
        # end if

        s = np.sum(change)

        if np.isnan(s):
            return 0
        # end if

        return abs(s / change.size)

    except (ValueError, ZeroDivisionError, IndexError):
        return 0
    # end try
# end average_shrink_rate

def down_movement(data: Iterable[float]) -> np.array:
    """
    Calculates the average change rate.

    :param data: The simulation data.

    :return: The calculated number.
    """

    data = array(data)

    validate_data_length(data)

    try:
        values = [0]

        for i, value in enumerate(data[1:], start=1):
            values.append(
                value - data[i - 1]
                if value < data[i - 1]
                else values[-1]
            )
        # end for

        values = array(values)

        return np.array(
            [
                np.sum(np.unique(values[:i + 1]))
                for i in range(values.size)
            ]
        )

    except (ValueError, ZeroDivisionError, IndexError):
        return 0
    # end try
# end down_movement

def up_movement(data: Iterable[float]) -> np.array:
    """
    Calculates the average change rate.

    :param data: The simulation data.

    :return: The calculated number.
    """

    data = array(data)

    validate_data_length(data)

    try:
        return -down_movement(data) + data

    except (ValueError, ZeroDivisionError):
        return 0
    # end try
# end up_movement

def cumulative_shrink(data: Iterable[float]) -> float:
    """
    Calculates the average change rate.

    :param data: The simulation data.

    :return: The calculated number.
    """

    data = array(data)

    validate_data_length(data)

    try:
        change = data[1:] - pd.Series(data).shift(1)[1:]

        return np.sum(np.array(np.where(change < 0, -change, 0))) / data[0]

    except (ValueError, ZeroDivisionError, IndexError):
        return 0
    # end try
# end cumulative_shrink

def cumulative_gain(data: Iterable[float]) -> float:
    """
    Calculates the average change rate.

    :param data: The simulation data.

    :return: The calculated number.
    """

    data = array(data)

    validate_data_length(data)

    try:
        change = data[1:] - pd.Series(data).shift(1)[1:]

        return np.sum(np.array(np.where(change > 0, change, 0))) / data[0]

    except (ValueError, ZeroDivisionError, IndexError):
        return 0
    # end try
# end cumulative_gain

def average_gain_rate(data: Iterable[float]) -> float:
    """
    Calculates the average change rate.

    :param data: The simulation data.

    :return: The calculated number.
    """

    data = array(data)

    validate_data_length(data)

    try:
        change = data[1:] - pd.Series(data).shift(1)[1:]

        change = np.where(change > 0, change, 0)
        change = np.array([np.sum(change[:i + 1]) for i in range(change.size)])
        change = change[1:] - pd.Series(change).shift(1)[1:]

        s = np.sum(change)

        if np.isnan(s):
            return 0
        # end if

        if change.size == 0:
            return 0
        # end if

        return abs(s / change.size)

    except (ValueError, ZeroDivisionError, IndexError):
        return 0
    # end try
# end average_gain_rate

def average_deviation_rate(data: Iterable[float]) -> float:
    """
    Calculates the average change rate.

    :param data: The simulation data.

    :return: The calculated number.
    """

    data = array(data)

    validate_data_length(data)

    try:
        change = np.abs(data[1:] - pd.Series(data).shift(1)[1:])

        s = np.sum(change)

        if np.isnan(s):
            return 0
        # end if

        return s / change.size

    except (ValueError, ZeroDivisionError, IndexError):
        return 0
    # end try
# end average_growth_rate

def min_change_factor(data: Iterable[float]) -> float:
    """
    Calculates the average change rate.

    :param data: The simulation data.

    :return: The calculated number.
    """

    data = array(data)

    validate_data_length(data)

    try:
        return np.min(data[1:] - pd.Series(data).shift(1)[1:])

    except (ValueError, ZeroDivisionError, IndexError):
        return 0
    # end try
# end min_change_factor

def max_change_factor(data: Iterable[float]) -> float:
    """
    Calculates the average change rate.

    :param data: The simulation data.

    :return: The calculated number.
    """

    data = array(data)

    validate_data_length(data)

    try:
        return np.max(data[1:] - pd.Series(data).shift(1)[1:])

    except (ValueError, ZeroDivisionError, IndexError):
        return 0
    # end try
# end max_change_factor

def average_change_factor(data: Iterable[float]) -> float:
    """
    Calculates the average change rate.

    :param data: The simulation data.

    :return: The calculated number.
    """

    data = array(data)

    validate_data_length(data)

    try:
        return np.sum(
            data[1:] - pd.Series(data).shift(1)[1:]
        ) / data[1:].size

    except (ValueError, ZeroDivisionError, IndexError):
        return 0
    # end try
# end average_change_factor

def average_win_rate(data: Iterable[float]) -> float:
    """
    Calculates the average change rate.

    :param data: The simulation data.

    :return: The calculated number.
    """

    data = array(data)

    validate_data_length(data)

    try:
        change = data[1:] - pd.Series(data).shift(1)[1:]

        values = np.where(change > 1, 1, 0)

        profit = np.count_nonzero(values)
        loss = values.size - profit

        if profit == 0:
            return 0.0
        # end if

        if profit + loss == 0:
            return 1.0
        # end if

        return profit / (profit + loss)

    except (ValueError, IndexError):
        return 0
    # end try
# end average_win_rate

def average_change_rate(data: Iterable[float]) -> float:
    """
    Calculates the average change rate.

    :param data: The simulation data.

    :return: The calculated number.
    """

    data = array(data)

    validate_data_length(data)

    try:
        change = data[1:] - pd.Series(data).shift(1)[1:]

        values = np.where(change == 0, 1, 0)

        constant = np.count_nonzero(values)
        changes = values.size - constant

        if changes == 0:
            return 0.0
        # end if

        if constant + changes == 0:
            return 1.0
        # end if

        return changes / (constant + changes)

    except (ValueError, IndexError):
        return 0
    # end try
# end average_change_rate

def weighted_average_win_rate(data: Iterable[float]) -> float:
    """
    Calculates the average change rate.

    :param data: The simulation data.

    :return: The calculated number.
    """

    data = array(data)

    validate_data_length(data)

    try:
        change = data[1:] - pd.Series(data).shift(1)[1:]

        profit = np.sum(np.where(change > 0, change, 0))
        loss = np.sum(np.where(change < 0, change, 0))

        if profit == 0:
            return 0.0
        # end if

        if profit + loss == 0:
            return 1.0
        # end if

        return profit / (profit + loss)

    except (ValueError, ZeroDivisionError, IndexError):
        return 0
    # end try
# end weighted_average_win_rate

def spread_deviation_risk_ratio(
        bids: Iterable[float],
        asks: Iterable[float],
        window: Optional[int] = None,
        accurate: Optional[bool] = True,
        weighted: Optional[bool] = True
) -> float:
    """
    Calculates the risk of an asset reversing more than the difference.

    :param bids: The bid prices for the asset.
    :param asks: The ask prices for the asset.
    :param window: The window of data points to calculate the risk with.
    :param weighted: The value to calculate the weighted risk.
    :param accurate: The value to calculate the accurate risk.

    :return: The rist ratio according to standard deviation.
    """

    bids = array(bids)
    asks = array(asks)

    window = window or min([bids.size, asks.size])

    if window == 0:
        raise ValueError("No data was given.")
    # end if

    if (window > bids.size) or (window > asks.size):
        raise ValueError(
            f"Bids length: {bids.size} and asks: {asks.size} "
            f"length don't match window the window: {window}."
        )
    # end if

    bids = bids[-window:]
    asks = asks[-window:]

    difference = (bids - asks) / asks

    asks_std = np.std(asks)
    bids_std = np.std(bids)

    deviation = (bids_std - asks_std) / asks_std

    if np.isnan(deviation):
        deviation = 0
    # end if

    if deviation <= 0:
        return deviation + 1
    # end if

    risk = float(np.sum(difference) / difference.size + deviation) / 2

    if accurate or weighted:
        win = np.where(bids > asks, bids - asks, 0)
        lose = np.where(bids <= asks, asks - bids, 0)

    else:
        win = None
        lose = None
    # end if

    accuracy = None

    if accurate:
        win_count = np.count_nonzero(win)
        lose_count = np.count_nonzero(lose)

        accuracy = win_count / (win_count + lose_count)
    # end if

    if weighted:
        win_sum = float(np.sum(win))
        lose_sum = float(np.sum(lose))

        weighted_accuracy = win_sum / (win_sum + lose_sum)
        accuracy = accuracy or weighted_accuracy
        accuracy = (accuracy + weighted_accuracy) / 2
    # end if

    if accuracy is not None:
        mistake = 1 - accuracy

    else:
        mistake = risk
    # end if

    return (risk + mistake) / 2
# end spread_deviation_risk_ratio


# interval.py

from typing import Tuple, Union, Dict, Any, ClassVar, Optional
import datetime as dt

from attrs import define

from represent import represent, Modifiers


__all__ = [
    "interval_amount",
    "interval_time",
    "interval_total_time",
    "parts_to_interval",
    "interval_duration",
    "interval_to_parts",
    "validate_interval_parts",
    "validate_interval_amount",
    "validate_interval",
    "validate_interval_duration",
    "INTERVALS",
    "SECONDS",
    "MINUTES",
    "MONTHS",
    "HOURS",
    "DAYS",
    "YEARS",
    "WEEKS",
    "Interval",
    "is_valid_interval"
]

SECONDS = "s"
MINUTES = "m"
MONTHS = "mo"
HOURS = "h"
DAYS = "d"
YEARS = "y"
WEEKS = "w"

INTERVALS = {
    SECONDS: dt.timedelta(seconds=1),
    MINUTES: dt.timedelta(minutes=1),
    HOURS: dt.timedelta(hours=1),
    DAYS: dt.timedelta(days=1),
    WEEKS: dt.timedelta(days=7),
    MONTHS: dt.timedelta(days=30),
    YEARS: dt.timedelta(days=365)
}

def validate_interval_duration(duration: Any) -> str:
    """
    Validates the interval part.

    :param duration: The value to validate.

    :return: The valid interval part.
    """

    if not isinstance(duration, str) or duration not in INTERVALS:
        raise ValueError(
            f"Interval duration must be a non-empty "
            f"string from the options: {', '.join(INTERVALS)}, "
            f"not: {duration}"
        )
    # end if

    return duration
# end validate_interval_duration

def validate_interval_amount(amount: Optional[Any] = None) -> int:
    """
    Validates the interval part.

    :param amount: The value to validate.

    :return: The valid interval part.
    """

    if amount is None:
        amount = 1

    elif not isinstance(amount, int) or (amount <= 0):
        raise ValueError(
            f"Interval amount must be a positive "
            f"integer, not: {amount}"
        )
    # end if

    return amount
# end validate_interval_amount

def validate_interval_parts(amount: Any, duration: Optional[Any] = None) -> Tuple[int, str]:
    """
    Validates the interval parts.

    :param amount: The increment value to validate.
    :param duration: The duration value to validate.

    :return: The valid interval parts.
    """

    duration = validate_interval_duration(duration)
    amount = validate_interval_amount(amount)

    return amount, duration
# end validate_interval_parts

def validate_interval(interval: str) -> str:
    """
    Validates the interval parts.

    :param interval: The interval to validate..

    :return: The valid interval parts.
    """

    amount, duration = interval_to_parts(interval)

    validate_interval_duration(duration)
    validate_interval_amount(amount)

    return interval
# end validate_interval

def is_valid_interval(interval: str) -> bool:
    """
    Validates the interval value.

    :param interval: The interval for the data.

    :return: The validates value.
    """

    try:
        validate_interval(interval)

        return True

    except ValueError:
        return False
    # end try
# end is_valid_interval

@define(slots=False, init=False, repr=False, eq=False, unsafe_hash=True)
@represent
class Interval:
    """
    A class to represent a trading pair.

    This object represents a pair of assets that can be traded.

    attributes:

    - base:
        The asset to buy or sell.

    - quote:
        The asset to use to buy or sell.

    >>> from market_break.interval import Interval
    >>>
    >>> one_day = Interval(1, "d")
    >>> five_seconds = Interval(5, "s")
    """

    __slots__ = "_amount", "_duration"

    __modifiers__ = Modifiers(excluded=["parts"])

    PERIODS: ClassVar[str] = 'periods'
    DURATION: ClassVar[str] = 'duration'

    try:
        from typing import Self

    except ImportError:
        Self = Any
    # end try

    def __init__(self, amount: int, duration: str) -> None:
        """
        Defines the class attributes.

        :param amount: The amount of periods for the interval.
        :param duration: The duration type for the interval.
        """

        amount, duration = validate_interval_parts(
            amount=amount, duration=duration
        )

        self._amount = amount
        self._duration = duration
    # end __init__

    def __eq__(self, other: Any) -> bool:
        """
        Checks if the signatures are equal.

        :param other: The signature to compare.

        :return: The equality value.
        """

        if type(other) is not type(self):
            return NotImplemented
        # end if

        other: Interval

        return (self is other) or (
            (self.amount == other.amount) and
            (self.duration == other.duration)
        )
    # end __eq__

    @property
    def amount(self) -> int:
        """
        Returns the periods property.

        :return: The amount of periods in the interval.
        """

        return self._amount
    # end amount

    @property
    def duration(self) -> str:
        """
        Returns the duration property.

        :return: The duration in the interval.
        """

        return self._duration
    # end duration

    @classmethod
    def load(cls, data: Dict[str, Union[str, int]]) -> Self:
        """
        Creates a pair of assets from the data.

        :param data: The pair data.

        :return: The pair object.
        """

        return cls(
            amount=data[cls.PERIODS],
            duration=data[cls.DURATION]
        )
    # end load

    @property
    def parts(self) -> Tuple[int, str]:
        """
        Returns the property value.

        :return: The symbol.
        """

        return self._amount, self._duration
    # end parts

    def interval(self) -> str:
        """
        Returns the string for the interval.

        :return: The string.
        """

        return f"{self.amount}{self.duration}"
    # end __str__

    def json(self) -> Dict[str, Union[int, str]]:
        """
        Converts the data into a json format.

        :return: The chain of assets.
        """

        return {
            self.PERIODS: self.amount,
            self.DURATION: self.duration
        }
    # end json
# end Interval

def interval_amount(interval: str) -> int:
    """
    Extracts the number from the interval.

    :param interval: The interval to extract.

    :return: The number from the interval.
    """

    for kind in sorted(INTERVALS.keys(), key=lambda key: len(key), reverse=True):
        try:
            if kind not in interval:
                continue
            # end if

            return int(interval.replace(kind, ""))

        except (TypeError, EOFError):
            pass
        # end try
    # end for

    raise ValueError(f"Invalid interval value: {interval}.")
# end interval_amount

def interval_duration(interval: str) -> str:
    """
    Extracts the type from the interval.

    :param interval: The interval to extract.

    :return: The type from the interval.
    """

    return interval.replace(str(interval_amount(interval)), "")
# end interval_duration

def interval_time(interval: str) -> dt.timedelta:
    """
    Extracts the type from the interval.

    :param interval: The interval to extract.

    :return: The type from the interval.
    """

    try:
        return INTERVALS[interval_duration(interval)]

    except KeyError:
        raise ValueError(f"Invalid interval structure: {interval}.")
    # end try
# end interval_time

def interval_total_time(interval: str) -> dt.timedelta:
    """
    Extracts the type from the interval.

    :param interval: The interval to extract.

    :return: The type from the interval.
    """

    return interval_amount(interval) * interval_time(interval)
# end interval_total_time

def parts_to_interval(increment: str, duration: int) -> str:
    """
    Creates a valid interval from the parameters.

    :param increment: The increment type for the interval.
    :param duration: The duration of the interval.

    :return: The interval.
    """

    if increment not in INTERVALS:
        raise ValueError(
            f"Interval increment must be one of "
            f"{', '.join(INTERVALS.keys())}, not {increment}."
        )
    # end if

    return f"{duration}{increment}"
# end parts_to_interval

def interval_to_parts(interval: str) -> Tuple[int, str]:
    """
    Creates a valid interval from the parameters.

    :param interval: The interval to separate to its parts.

    :return: The interval parts.
    """

    amount = interval_amount(interval)

    return amount, interval.replace(str(amount), "")
# end parts_to_interval


# evaluation.py

import json
from typing import Iterable, Optional, Union, Any, Dict, List

import numpy as np

from represent import represent, Modifiers

from market_break.process import (
    average_growth_rate,
    average_win_rate, average_shrink_rate, array,
    average_gain_rate, weighted_average_win_rate,
    average_change_rate, cumulative_shrink, cumulative_gain
)

__all__ = [
    "EvaluationResults",
    "results_report",
    "EvaluationResultsRecorder"
]

@represent
class EvaluationResults:
    """
    A class to represent a simulation results.
    This project is returned from the simulation function, as a
    data container to contain the simulation results, as well as
    the important initial parameters and values.

    This object can also be used to run analysis on the results' data,
    using the methods of the class.

    attributes:

    - balances:
        The balance values that were calculated during the simulation process.

    - predictions:
        The predicted balance values that were calculated during the simulation process.

    - trader:
        The trader object for the origin of the data.

    - best:
        The value to execute only the best arbitrage chain when several are found.

    - index:
        The index values of thew simulation process.
    """

    __modifiers__ = Modifiers(hidden=["predictions", "balances", "index"])

    def __init__(
            self,
            balances: Optional[Iterable[float]] = None,
            predictions: Optional[Iterable[float]] = None,
            index: Optional[Iterable] = None
    ) -> None:
        """
        Defines the class attributes.

        :param balances: The balances of the trades.
        :param predictions: The predictions of the trades.
        :param index: The index for the process.
        """

        self._balance_growth_rate = 0
        self._average_win_rate = 0
        self._weighted_average_win_rate = 0
        self._activity_rate = 0
        self._deviation_rate = 0
        self._prediction_accuracy = 0
        self._loss_growth_rate = 0
        self._profit_growth_rate = 0
        self._cumulative_shrink = 0
        self._cumulative_gain = 0

        self.balances = array(balances or [])
        self.predictions = array(predictions or [])
        self.index = index or []
    # end __init__

    @property
    def balance(self) -> float:
        """
        Returns thew value of the results.

        :return: The value.
        """

        if len(self.balances) > 0:
            return self.balances[-1] / self.balances[0]

        else:
            return 0
        # end if
    # end balance

    @property
    def profit(self) -> float:
        """
        Returns thew value of the results.

        :return: The value.
        """

        if len(self.balances) > 0:
            return self.balance - 1 if self.balance > 1 else 0

        else:
            return 0
        # end if
    # end profit

    @property
    def loss(self) -> float:
        """
        Returns thew value of the results.

        :return: The value.
        """

        if len(self.balances) > 0:
            return self.balance - 1 if self.balance < 1 else 0

        else:
            return 0
        # end if
    # end profit

    def cumulative_shrink(self) -> float:
        """
        Calculates the average balance rate.

        :return: The calculated number.
        """

        if (len(self.balances) > 1) and (self._cumulative_shrink == 0):
            gross_loss = cumulative_shrink(self.balances)

            if np.isnan(gross_loss):
                self._cumulative_shrink = 0

            else:
                self._cumulative_shrink = gross_loss
            # end if
        # end if

        return self._cumulative_shrink
    # end cumulative_shrink

    def cumulative_gain(self) -> float:
        """
        Calculates the average balance rate.

        :return: The calculated number.
        """

        if (len(self.balances) > 1) and (self._cumulative_gain == 0):
            gross_profit = cumulative_gain(self.balances)

            if np.isnan(gross_profit):
                self._cumulative_gain = 0

            else:
                self._cumulative_gain = gross_profit
            # end if
        # end if

        return self._cumulative_gain
    # end cumulative_gain

    def loss_growth_rate(self) -> float:
        """
        Calculates the average balance rate.

        :return: The calculated number.
        """

        if (len(self.balances) > 1) and (self._loss_growth_rate == 0):
            loss_growth_rate = average_shrink_rate(self.balances)

            if np.isnan(loss_growth_rate):
                self._loss_growth_rate = 0

            else:
                self._loss_growth_rate = loss_growth_rate
            # end if
        # end if

        return self._loss_growth_rate
    # end loss_growth_rate

    def profit_growth_rate(self) -> float:
        """
        Calculates the average balance rate.

        :return: The calculated number.
        """

        if (len(self.balances) > 1) and (self._profit_growth_rate == 0):
            profit_growth_rate = average_gain_rate(self.balances)

            if np.isnan(profit_growth_rate):
                self._profit_growth_rate = 0

            else:
                self._profit_growth_rate = profit_growth_rate
            # end if
        # end if

        return self._profit_growth_rate
    # end profit_growth_rate

    def balance_growth_rate(self) -> float:
        """
        Calculates the average balance rate.

        :return: The calculated number.
        """

        if (len(self.balances) > 1) and (self._balance_growth_rate == 0):
            balance_growth_rate = average_growth_rate(self.balances)

            if np.isnan(balance_growth_rate):
                self._balance_growth_rate = 0

            else:
                self._balance_growth_rate = balance_growth_rate
            # end if
        # end if

        return self._balance_growth_rate
    # end balance_growth_rate

    def win_rate(self) -> float:
        """
        Calculates the average win rate.

        :return: The calculated number.
        """

        if (len(self.balances) > 1) and (self._average_win_rate == 0):
            win_rate = average_win_rate(self.balances)

            if np.isnan(win_rate):
                self._average_win_rate = 0

            else:
                self._average_win_rate = win_rate
            # end if
        # end if

        return self._average_win_rate
    # end win_rate

    def weighted_win_rate(self) -> float:
        """
        Calculates the average win rate.

        :return: The calculated number.
        """

        if (len(self.balances) > 1) and (self._weighted_average_win_rate == 0):
            weighted_win_rate = weighted_average_win_rate(self.balances)

            if np.isnan(weighted_win_rate):
                self._weighted_average_win_rate = 0

            else:
                self._weighted_average_win_rate = weighted_win_rate
            # end if
        # end if

        return self._weighted_average_win_rate
    # end win_rate

    def activity_rate(self) -> float:
        """
        Calculates the average win rate.

        :return: The calculated number.
        """

        if (len(self.balances) > 1) and (self._activity_rate == 0):
            activity_rate = len(np.unique(self.balances)) / len(self.balances)

            if np.isnan(activity_rate):
                self._activity_rate = 0

            else:
                self._activity_rate = activity_rate
            # end if
        # end if

        return self._activity_rate
    # end activity_rate

    def deviation_rate(self) -> float:
        """
        Calculates the average win rate.

        :return: The calculated number.
        """

        if (len(self.balances) > 1) and (self._deviation_rate == 0):
            deviation_rate = average_change_rate(self.balances)

            if np.isnan(deviation_rate):
                self._deviation_rate = 0

            else:
                self._deviation_rate = deviation_rate
            # end if
        # end if

        return self._deviation_rate
    # end deviation_rate

    def prediction_accuracy(self) -> float:
        """
        Calculates the average win rate.

        :return: The calculated number.
        """

        if (self.predictions is None) or (len(self.predictions) != len(self.balances)):
            return np.nan
        # end if

        if (len(self.balances) > 1) and (self._prediction_accuracy == 0):
            predictions_change = (
                np.abs(self.predictions - self.balances) /
                self.balances
            )

            prediction_accuracy = (
                1 - (
                    np.sum(predictions_change) /
                    len(predictions_change)
                )
            )

            if np.isnan(prediction_accuracy):
                self._average_win_rate = 0

            else:
                self._prediction_accuracy = prediction_accuracy
            # end if
        # end if

        return self._prediction_accuracy
    # end prediction_accuracy
# end EvaluationResults

class EvaluationResultsRecorder:
    """A class to represent a record visualizer."""

    INDEX = "index"
    BALANCES = "balances"
    ADDING_BALANCES = "adding_balances"
    PREDICTIONS = "predictions"
    ADDING_PREDICTIONS = "predictions"

    ATTRIBUTES = (INDEX, BALANCES, ADDING_BALANCES, PREDICTIONS, ADDING_PREDICTIONS)

    def __init__(self, keys: Iterable[Any]) -> None:
        """
        Defines the class attributes.

        :param keys: The keys for the reports.
        """

        self.record: Dict[str, Dict[str, List[Union[float, Any]]]] = {
            key: {attribute: [] for attribute in self.ATTRIBUTES}
            for key in keys
        }

        self.sum_record: Dict[str, List[Union[float, Any]]] = {
            attribute: [] for attribute in self.ATTRIBUTES
        }
    # end __init__

    def update(
            self,
            data: Dict[Any, Union[Dict[str, Iterable[Union[float, Any]]], EvaluationResults]],
            start: Optional[float] = None,
            adjust: Optional[bool] = True,
            sort: Optional[bool] = True,
            padding: Optional[bool] = False
    ) -> None:
        """
        Updates the record of the evaluation results.

        :param data: The data to add.
        :param start: start value for the adding values.
        :param adjust: The value to adjust for unknown keys.
        :param sort: The value to sort the indexes.
        :param padding: The value to pad the short values.
        """

        if start is None:
            start = 0
        # end if

        add = True

        start_add_balance = (
            self.sum_record[self.ADDING_BALANCES][0]
            if self.sum_record[self.ADDING_BALANCES] else start
        )
        start_add_prediction = (
            self.sum_record[self.ADDING_PREDICTIONS][0]
            if self.sum_record[self.ADDING_PREDICTIONS] else start
        )

        for key, results in data.items():
            if key not in self.record:
                if not adjust:
                    raise KeyError(
                        f"Key: {repr(key)} is not present in the record. "
                        f"All keys must be defined during initialization."
                    )

                else:
                    continue
                # end if
            # end if

            for attribute in (self.INDEX, self.BALANCES, self.PREDICTIONS):
                if (
                    hasattr(results, attribute)
                    if isinstance(results, EvaluationResults) else
                    attribute in results
                ):
                    # noinspection PyUnresolvedReferences
                    for i, value in enumerate(
                        getattr(results, attribute)
                        if isinstance(results, EvaluationResults) else
                        results[attribute]
                    ):
                        self.record[key][attribute].append(value)

                        if attribute != self.INDEX:
                            self.record[key][f"adding_{attribute}"].append(
                                value + self.record[key][attribute][-1]
                            )

                            if add:
                                self.sum_record[attribute].append(value)
                                self.sum_record[f"adding_{attribute}"].append(
                                    value + (
                                        self.sum_record[f"adding_{attribute}"][-1]
                                        if self.sum_record[f"adding_{attribute}"] else
                                        self.sum_record[attribute][-1]
                                    )
                                )

                                add = False

                            else:
                                self.sum_record[attribute][i] += value
                                self.sum_record[f"adding_{attribute}"][i] += value
                            # end if
                        # end if

                        if (
                            (attribute == self.INDEX) and
                            (value not in self.sum_record[attribute])
                        ):
                            self.sum_record[attribute].append(value)

                            if sort:
                                try:
                                    self.sum_record[attribute].sort()

                                except Exception as e:
                                    if adjust:
                                        pass

                                    else:
                                        raise e
                                    # end if
                                # end try
                            # end if
                        # end if
                    # end for
                # end if
            # end for
        # end for

        try:
            self.sum_record[self.ADDING_BALANCES][0] = start_add_balance
            self.sum_record[self.ADDING_PREDICTIONS][0] = start_add_prediction

        except IndexError:
            pass
        # end try

        if padding:
            length = max(
                max(len(values) for values in result.values())
                for result in self.record.values()
            )

            for key, value in self.record.items():
                self.record[key] = {
                    attribute: [*values, *([0] * (length - len(values)))]
                    for attribute, values in value.items()
                }
            # end for
        # end if
    # end update

    def sum_evaluation_result(self) -> EvaluationResults:
        """
        Creates all results from the record.

        :return: The result objects.
        """

        return EvaluationResults(
            **{
                attribute: values
                for attribute, values in self.sum_record.items()
                if attribute in (self.INDEX, self.BALANCES, self.PREDICTIONS)
            }
        )
    # end sum_evaluation_result

    def adding_sum_evaluation_result(self) -> EvaluationResults:
        """
        Creates all results from the record.

        :return: The result objects.
        """

        return EvaluationResults(
            **{
                attribute.replace("adding_", ""): values
                for attribute, values in self.sum_record.items()
                if attribute in (
                    self.INDEX, self.ADDING_BALANCES, self.ADDING_PREDICTIONS
                )
            }
        )
    # end adding_sum_evaluation_result

    def evaluation_results(
        self, padding: Optional[bool] = False
    ) -> Dict[Any, EvaluationResults]:
        """
        Creates all results from the record.

        :param padding: The value to pad the short values.

        :return: The result objects.
        """

        length = max(
            max(len(v) for v in result.values())
            for result in self.record.values()
        )

        return {
            key: EvaluationResults(
                **{
                    attribute: (
                        [*([0] * (length - len(values))), *values]
                        if padding else values
                    )
                    for attribute, values in results.items()
                    if attribute in (self.INDEX, self.BALANCES, self.PREDICTIONS)
                }
            ) for key, results in self.record.items()
        }
    # end evaluation_results

    def adding_evaluation_results(
            self, padding: Optional[bool] = False
    ) -> Dict[Any, EvaluationResults]:
        """
        Creates all results from the record.

        :param padding: The value to pad the short values.

        :return: The result objects.
        """

        length = max(
            max(len(v) for v in result.values())
            for result in self.record.values()
        )

        return {
            key: EvaluationResults(
                **{
                    attribute.replace("adding_", ""): (
                        [*([0] * (length - len(values))), *values]
                        if padding else values
                    )
                    for attribute, values in results.items()
                    if attribute in (
                        self.INDEX, self.ADDING_BALANCES, self.ADDING_PREDICTIONS
                    )
                }
            ) for key, results in self.record.items()
        }
    # end adding_evaluation_results
# end EvaluationResultsRecorder

class JSONEncoder(json.JSONEncoder):
    """A class to represent a json encoder."""

    def default(self, obj: Any) -> Any:
        """
        encodes the value of the object to valid json data.

        :param obj: The object to encode.

        :return: The encoded json value.
        """

        if isinstance(obj, np.integer):
            return int(obj)
        # end if

        if isinstance(obj, np.floating):
            return float(obj)
        # end if

        if isinstance(obj, np.ndarray):
            return obj.tolist()
        # end if

        try:
            if np.isnan(obj):
                obj = None
            # end if

        except TypeError:
            pass
        # end try

        return json.JSONEncoder.default(self, obj)
    # end default
# end JSONEncoder

def results_report(
        results: EvaluationResults,
        precision: Optional[int] = 3,
        prefix: Optional[str] = "",
        info: Optional[bool] = True,
        performance: Optional[bool] = True,
        total: Optional[bool] = True
) -> Dict[str, Union[str, float]]:
    """
    Calculates the results.

    :param results: The results object.
    :param precision: The precision of the floating point numbers.
    :param prefix: The prefix for each key.
    :param info: The value to include the basic info.
    :param performance: The value to include the performance analysis.
    :param total: The value to include the total results.

    :return: The data of the results.
    """

    data = {
        **(
            {
                f"{prefix}Info": "",
                f"{prefix}iterations": len(results.balances),
                f"{prefix}periods": len(results.index),
                **(
                    {
                        f"{prefix}start": str(results.index[0]),
                        f"{prefix}end": str(results.index[-1]),
                        f"{prefix}span": str(results.index[-1] - results.index[0]),
                    } if results.index else {}
                ),
                f"{prefix}starting balance": (
                    str(round(results.balances[0], precision))
                    if len(results.balances) > 0 else "none"
                ),
            } if info else {}
        ),
        **(
            {
                f"{prefix}Performance": "",
                f"{prefix}win rate (%)": results.win_rate() * 100,
                f"{prefix}weighted win rate (%)": results.weighted_win_rate() * 100,
                f"{prefix}activity rate (%)": results.activity_rate() * 100,
                f"{prefix}deviation rate (%)": results.deviation_rate() * 100,
                f"{prefix}accuracy rate (%)": results.prediction_accuracy() * 100,
                f"{prefix}profit growth rate (%)": results.profit_growth_rate() * 100,
                f"{prefix}loss growth rate (%)": results.loss_growth_rate() * 100,
                f"{prefix}balance growth rate (%)": results.balance_growth_rate() * 100,
            } if performance else {}
        ),
        **(
            {
                f"{prefix}Total Results": "",
                f"{prefix}ending balance": (
                    str(round(results.balances[-1], precision))
                    if len(results.balances) > 0 else "none"
                ),
                f"{prefix}gross profit (%)": results.cumulative_gain() * 100,
                f"{prefix}net profit (%)": results.profit * 100,
                f"{prefix}gross loss (%)": results.cumulative_shrink() * 100,
                f"{prefix}net loss (%)": results.loss * 100,
                f"{prefix}balance (%)": results.balance * 100
            } if total else {}
        )
    }

    return json.loads(json.dumps(data, cls=JSONEncoder))
# end results_report

def bidirectional_results_report(
        long_results: EvaluationResults,
        short_results: EvaluationResults,
        precision: Optional[int] = 3,
        prefix: Optional[str] = "",
        info: Optional[bool] = True,
        performance: Optional[bool] = True,
        total: Optional[bool] = True
) -> Dict[str, Union[str, float]]:
    """
    Calculates the results.

    :param long_results: The long results object.
    :param short_results: The long results object.
    :param precision: The precision of the floating point numbers.
    :param prefix: The prefix for each key.
    :param info: The value to include the basic info.
    :param performance: The value to include the performance analysis.
    :param total: The value to include the total results.

    :return: The data of the results.
    """

    results = EvaluationResults(
        index=long_results.index,
        balances=long_results.balances + short_results.balances
    )

    data = {
        **results_report(
            results=results, precision=precision, prefix=prefix,
            info=info, performance=performance, total=total
        ),
        **results_report(
            results=long_results, precision=precision,
            prefix=prefix, performance=performance, total=total
        ),
        **results_report(
            results=short_results, precision=precision,
            prefix=prefix, performance=performance, total=total
        )
    }

    return json.loads(json.dumps(data, cls=JSONEncoder))
# end bidirectional_results_report